% This code accompanies the paper:
%
% "Elastic Net Constraints for Shape Matching"
% E.Rodola, A.Torsello, T.Harada, Y.Kuniyoshi, and D.Cremers
% 2013 IEEE International Conference on Computer Vision (ICCV 2013)
%
% Please cite the paper above if you use this code.
%
% Emanuele Rodola
% The University of Tokyo, 2012
% TU Munich, 2013
% <rodola@in.tum.de>
%
function [x, iters, err] = enet_grad_desc(S, alpha, x_init, accuracy, max_iters, step)
% ENET_GRAD_DESC  Projected gradient descent solution of a QAP under 
%                 elastic net constraints.
%
%   Usage:
%     [X, I, E] = ENET_GRAD_DESC(S, ALPHA)
%     [X, I, E] = ENET_GRAD_DESC(S, ALPHA, X_INIT)
%     [X, I, E] = ENET_GRAD_DESC(S, ALPHA, X_INIT, ACCURACY)
%     [X, I, E] = ENET_GRAD_DESC(S, ALPHA, X_INIT, ACCURACY, MAX_ITERS)
%     [X, I, E] = ENET_GRAD_DESC(S, ALPHA, X_INIT, ACCURACY, MAX_ITERS, STEP)
%
%   Input:
%     S         Similarity matrix for the QAP.
%     ALPHA     Convexity factor for the elastic net ball. ALPHA=0 
%               corresponds to the L1 ball, ALPHA=1 corresponds to the L2 
%               ball. Any value within (0,1) can be passed.
%     X_INIT    Starting point for the gradient descent procedure. Default
%               starts from the barycenter of the elastic net boundary.
%     ACCURACY  Desired final accuracy (relative error). (default: 1e-6).
%     MAX_ITERS Maximum number of iterations. (default: 1e4).
%     STEP      Step length for gradient descent. (default: 1e-4).
%
%   Output:
%     X     Final solution of the enet-regularized QAP.
%     ITERS Final number of iterations taken by the gradient descent.
%     ERR   Final error attained by the gradient descent process.
%
%   Note:
%     This code is solving a *maximization* problem. If you want to solve a
%     minimization problem instead, you can pass -C in place of S, where C
%     is your cost matrix. Alternatively, you can modify the code below to
%     go in the direction of the negative gradient and use C directly.
%
%   Author: Emanuele Rodola (c) 2014
%           TU Munich, rodola@in.tum.de

if nargin < 2
    error('At least the similarity matrix and convexity coefficient must be specified.');
end

if size(S,1) ~= size(S,2)
    error('The similarity matrix must be a square matrix.');
end

if alpha<=0 || alpha>=1
    error('The convexity parameter for the elastic net ball must be in (0,1).');
end

if nargin < 6
    step = 1e-4;
end

if nargin < 5
    max_iters = 1e4;
end

if nargin < 4
    accuracy = 1e-6;
end

if (nargin < 3) || isempty(x_init)
    n = size(S,1);
    x_init = ((n*(alpha-1) + (((n-n*alpha)^2+4*alpha*n)^0.5)) / (2*alpha*n)) * ones(n,1);
else
    x_init = proj_enet(x_init, alpha);
end

x = x_init;
xaxs = zeros(1,max_iters);

for iters=1:max_iters
    
    Sx = S*x;
    
    x = x + step*Sx;
    [x,~] = proj_enet(x, alpha, 1);
    
    xaxs(iters) = x'*Sx;
    
    if iters>1
        err = xaxs(iters) - xaxs(iters-1);
        if err < 0
            warning('Machine precision reached (iteration %d).',iters);
            break;
        end
        if err < accuracy
            break;
        end
    end
    
    if iters==max_iters
        break;
    end
    
end

end
