/*
 * This code accompanies the paper:
 *
 * "Elastic Net Constraints for Shape Matching"
 * E.Rodola, A.Torsello, T.Harada, Y.Kuniyoshi, and D.Cremers
 * 2013 IEEE International Conference on Computer Vision (ICCV 2013)
 *
 * Please cite the paper above if you use this code.
 *
 * Emanuele Rodola
 * The University of Tokyo, 2012
 * TU Munich, 2013
 * <rodola@in.tum.de>
 */

#include "mex.h"
#include <cmath>
#include <vector>

 /**
  * Performs exact projection of a given vector onto the elastic net ball
  * with given convexity 'a' and radius 't'.
  *
  * NOTE: 
  * All for-loops in the project() function can be made trivially parallel,
  * e.g. by compiling with OpenMP and using #pragma omp parallel for
  *
  * @param x     Input vector to project onto the elastic net ball.
  * @param n     Size of the input vector. Must be > 0.
  * @param a     Convexity of the elastic net ball. Must be in (0,1).
  * @param t     Radius of the elastic net ball. Must be > 0.
  * @param proj  Output projected vector.
  * @param iters Output final number of iterations of the projection process.
  */
void project(const double* x, int n, double a, double t, double* proj, int* iters)
{
	*iters = 0;
	
	const double ac = 1-a;
	const double acsq = ac*ac;
	
	double pb = 2*t*a + (n/2.)*acsq;
	
	double pc = 0.;
	for (int i=0; i<n; ++i)
	{
		const double xi = x[i];
		pc += a*xi*xi + ac*xi;
	}

	if (pc < t) // already inside the ball
	{
		for (int i=0; i<n; ++i)
			proj[i] = x[i];
		return;
	}
	
	pc -= t;
	
	double lambda = (pb - std::sqrt(pb*pb + 2*a*pb*pc)) / (-a*pb);
	
	std::vector<int> e(n,true);
	
	int n_neg = 0;
	double num = lambda*ac/2;
	double den = 1+lambda*a;
	for (int i=0; i<n; ++i)
	{
		proj[i] = (x[i] - num) / den;
		if (proj[i] < 0)
		{
			e[i] = false;
			++n_neg;
		}
	}
    
	double apb;
	bool done = (n_neg == 0);

    while (!done) // this external loop cannot be made parallel
	{
		++(*iters);
		
		done = true;
		
		pb = 2*t*a + ((n-n_neg)/2.)*acsq;
		apb = a*pb;
		
		pc = 0.;
		for (int i=0; i<n; ++i)
		{
			if (e[i])
			{
				const double xi = x[i];
				pc += a*xi*xi + ac*xi;
			}
		}
		pc -= t;
		
        lambda = (pb - std::sqrt(pb*pb + 2*apb*pc)) / (-apb);

		num = lambda*ac/2;
		den = 1+lambda*a;
		for (int i=0; i<n; ++i)
		{
			if (e[i])
			{
				proj[i] = (x[i] - num) / den;
				if (proj[i] < 0)
				{
					done = false;
					e[i] = false;
					++n_neg;
				}
			}
			else if (done)
				proj[i] = 0.;
		}
    } // next iteration
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray*prhs[])
{
	if (nrhs != 3 || nlhs != 2)
      mexErrMsgTxt("Usage: [proj,iters] = proj_enet(x, alpha, radius).");

	const double* const x = mxGetPr(prhs[0]);
	const int m = int( mxGetM(prhs[0]) );
	const int n = int( mxGetN(prhs[0]) );
	
	if ((m!=1 && n!=1) || (m==0 || n==0))
		mexErrMsgTxt("The first argument must be either a row or column vector.");
		
	const int l = (m > n ? m : n);
	
	const double alpha = *(double*)mxGetData(prhs[1]);
	
	if (alpha <= 0. || alpha >= 1.)
		mexErrMsgTxt("The second argument must be a real scalar in (0,1).");
		
	const double radius = *(double*)mxGetData(prhs[2]);
	
	if (radius <= 0.)
		mexErrMsgTxt("The third argument must be a positive real scalar.");
	
	plhs[0] = mxCreateDoubleMatrix(l, 1, mxREAL);
    double* proj = mxGetPr(plhs[0]);
	
	int iters = 0;
	project(x, l, alpha, radius, proj, &iters);
	
	plhs[1] = mxCreateDoubleScalar(static_cast<double>(iters));
}
